#ifndef _X86_RWSEM_XCHGADD_H
#define _X86_RWSEM_XCHGADD_H

static inline void __down_read(struct rw_semaphore *sem)
{
	__asm__ __volatile__(LOCK "incl %0\n\t"
			     "js 2f\n"
			     "1:\n"
			     ".section .text.lock,\"ax\"\n" 
			     "2:\t"
			     "pushl %%edx\n\t"
			     "pushl %%ecx\n\t"
			     "movl %2, %%edx\n\t"
			     "call rwsem_down_failed\n\t"
			     "popl %%ecx\n\t"
			     "popl %%edx\n\t"
			     "jmp 1b\n"
			     ".previous"
			     : "+m" (sem->count)
			     : "a" (sem), "i" (RWSEM_READ_BLOCKING_BIAS)
			     : "memory", "cc");
}

static inline void __down_write(struct rw_semaphore *sem)
{
	long count;

	count = RWSEM_WRITE_BIAS + RWSEM_READ_BIAS;
	__asm__ __volatile(LOCK "xaddl %0, %1\n\t"
			   "testl %0,%0\n\t"
			   "jnz 2f\n"
			   "1:\n"
			   ".section .text.lock,\"ax\"\n"
			   "2:\t"
			   "pushl %%ecx\n\t"
			   "movl %3, %%edx\n\t"
			   "call rwsem_down_failed\n\t"
			   "popl %%ecx\n\t"
			   "jmp 1b\n"
			   ".previous"
			   : "+d" (count), "+m" (sem->count)
			   : "a" (sem), "i" (RWSEM_WRITE_BLOCKING_BIAS)
			   : "memory", "cc");
}

static inline void __up_read(struct rw_semaphore *sem)
{
	long count;

	count = -RWSEM_READ_BIAS;
	__asm__ __volatile__(LOCK "xaddl %0, %1\n\t"
			     "js 2f\n"
			     "1:\n"
			     ".section .text.lock,\"ax\"\n"
			     "2:\t"
			     "cmpw $1, %w0\n\t"
			     "jnz 1b\n\t"
			     "pushl %%ecx\n\t"
			     "call rwsem_wake\n\t"
			     "popl %%ecx\n\t"
			     "jmp 1b\n"
			     ".previous"
			     : "+d" (count), "+m" (sem->count)
			     : "a" (sem)
			     : "memory", "cc");
}
static inline void __up_write(struct rw_semaphore *sem)
{
	__asm__ __volatile__(LOCK "subl %2, %0\n\t"
			     "js 2f\n"
			     "1:\n"
			     ".section .text.lock,\"ax\"\n"
			     "2:\t"
			     "pushl %%edx\n\t"
			     "pushl %%ecx\n\t"
			     "call rwsem_wake\n\t"
			     "popl %%ecx\n\t"
			     "popl %%edx\n\t"
			     "jmp 1b\n"
			     ".previous"
			     : "+m" (sem->count)
			     : "a" (sem), "i" (RWSEM_READ_BIAS + RWSEM_WRITE_BIAS)
			     : "memory", "cc");
}

static inline long rwsem_xchgadd(long value, long * count)
{
	__asm__ __volatile__(LOCK "xaddl %0,%1"
			     : "+r" (value), "+m" (*count));
	return value;
}

#endif
