#ifndef _LINUX_RWSEM_SPINLOCK_H
#define _LINUX_RWSEM_SPINLOCK_H

#include <linux/compiler.h>
#include <linux/kernel.h>

struct rw_semaphore
{
	spinlock_t lock;
	long count;
#define RWSEM_READ_BIAS 1
#define RWSEM_WRITE_BIAS (~(~0UL >> (BITS_PER_LONG>>1)))
	struct list_head wait;
#if RWSEM_DEBUG
	long __magic;
#endif
};

#if RWSEM_DEBUG
#define __SEM_DEBUG_INIT(name) \
	, (long)&(name).__magic
#define RWSEM_MAGIC(x)							\
	do {								\
		if ((x) != (long)&(x)) {				\
			printk("rwsem bad magic %lx (should be %lx), ",	\
				(long)x, (long)&(x));			\
			BUG();						\
		}							\
	} while (0)
#else
#define __SEM_DEBUG_INIT(name)
#define CHECK_MAGIC(x)
#endif

#define __RWSEM_INITIALIZER(name, count)	\
{						\
	SPIN_LOCK_UNLOCKED,			\
	(count),				\
	LIST_HEAD_INIT((name).wait)		\
	__SEM_DEBUG_INIT(name)			\
}
#define RWSEM_INITIALIZER(name) __RWSEM_INITIALIZER(name, 0)

#define __DECLARE_RWSEM(name, count) \
	struct rw_semaphore name = __RWSEM_INITIALIZER(name, count)
#define DECLARE_RWSEM(name) __DECLARE_RWSEM(name, 0)
#define DECLARE_RWSEM_READ_LOCKED(name) __DECLARE_RWSEM(name, RWSEM_READ_BIAS)
#define DECLARE_RWSEM_WRITE_LOCKED(name) __DECLARE_RWSEM(name, RWSEM_WRITE_BIAS)

#define RWSEM_READ_BLOCKING_BIAS (RWSEM_WRITE_BIAS-RWSEM_READ_BIAS)
#define RWSEM_WRITE_BLOCKING_BIAS (0)

#define RWSEM_READ_MASK (~RWSEM_WRITE_BIAS)
#define RWSEM_WRITE_MASK (RWSEM_WRITE_BIAS)

extern void FASTCALL(init_rwsem(struct rw_semaphore *));
extern void FASTCALL(down_read(struct rw_semaphore *));
extern void FASTCALL(down_write(struct rw_semaphore *));
extern void FASTCALL(up_read(struct rw_semaphore *));
extern void FASTCALL(up_write(struct rw_semaphore *));

#endif /* _LINUX_RWSEM_SPINLOCK_H */
