/*
 *  rw_semaphores xchgadd version
 *  Copyright (C) 2001 Andrea Arcangeli <andrea@suse.de> SuSE
 */

#include <linux/sched.h>
#include <linux/module.h>
#include <asm/semaphore.h>

struct rwsem_wait_queue {
	unsigned long retire;
	struct task_struct * task;
	struct list_head task_list;
};

static void FASTCALL(__rwsem_wake(struct rw_semaphore *));
static void __rwsem_wake(struct rw_semaphore *sem)
{
	struct list_head * entry, * head = &sem->wait;
	int wake_write = 0, wake_read = 0;

	while ((entry = head->prev) != head) {
		struct rwsem_wait_queue * wait;
		long count;

		wait = list_entry(entry, struct rwsem_wait_queue, task_list);

		if (wait->retire == RWSEM_WRITE_BLOCKING_BIAS) {
			if (wake_read)
				break;
			wake_write = 1;
		}

	again:
		count = rwsem_xchgadd(-wait->retire, &sem->count);
		if (!wake_read && (count & RWSEM_READ_MASK)) {
			count = rwsem_xchgadd(wait->retire, &sem->count);
			if ((count & RWSEM_READ_MASK) == 1)
				goto again;
			break;
		}
		
		list_del(entry);
		entry->next = NULL;
		wake_up_process(wait->task);
			
		if (wake_write)
			break;
		wake_read = 1;
	}
}

struct rw_semaphore * rwsem_down_failed(struct rw_semaphore *sem, long retire)
{
	struct task_struct *tsk = current;
	struct rwsem_wait_queue wait;
	long count;

	wait.retire = retire;
	wait.task = tsk;
	INIT_LIST_HEAD(&wait.task_list);

	spin_lock(&sem->lock);
	list_add(&wait.task_list, &sem->wait);

	count = rwsem_xchgadd(retire, &sem->count);
	if ((count & RWSEM_READ_MASK) == 1)
		__rwsem_wake(sem);

	while (wait.task_list.next) {
		__set_task_state(tsk, TASK_UNINTERRUPTIBLE);
		spin_unlock(&sem->lock);
		schedule();
		spin_lock(&sem->lock);
	}

	spin_unlock(&sem->lock);

	return sem;
}

struct rw_semaphore * rwsem_wake(struct rw_semaphore *sem)
{
	spin_lock(&sem->lock);
	__rwsem_wake(sem);
	spin_unlock(&sem->lock);

	return sem;
}

EXPORT_SYMBOL_NOVERS(rwsem_down_failed);
EXPORT_SYMBOL_NOVERS(rwsem_wake);
